"use strict";
var _a, _b;
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const prompts_1 = require("@inquirer/prompts");
const config_1 = require("../lib/config");
const os_1 = require("os");
const core_1 = require("@oclif/core");
const chalk_1 = require("chalk");
const node_fetch_1 = tslib_1.__importDefault(require("node-fetch"));
const secret_1 = require("../lib/secret");
const DEFAULT_NAME = "cloud";
const DB = (_a = process.env.FAUNA_URL) !== null && _a !== void 0 ? _a : "https://db.fauna.com";
const AUTH = (_b = process.env.FAUNA_AUTH) !== null && _b !== void 0 ? _b : "https://auth.console.fauna.com";
class Region {
    constructor(name, secret) {
        this.name = name;
        this.secret = secret;
    }
    endpointName(base) {
        return `${base}-${this.name}`;
    }
}
class CloudLoginCommand extends core_1.Command {
    async run() {
        const config = config_1.ShellConfig.read({});
        await this.execute(config);
    }
    async execute(config) {
        await this.parse();
        const base = await this.askName(config);
        const regions = await this.passwordStrategy();
        const newDefault = await this.askDefault(config, base, regions);
        for (const region of Object.values(regions)) {
            config.rootConfig.endpoints[region.endpointName(base)] = new config_1.Endpoint({
                url: DB,
                secret: secret_1.Secret.parse(region.secret),
            });
        }
        config.rootConfig.defaultEndpoint = newDefault;
        config.saveRootConfig();
        this.log("Configuration updated.");
    }
    async askName(config) {
        const name = await (0, prompts_1.input)({
            message: "Endpoint name",
            default: DEFAULT_NAME,
            validate: (endpoint) => endpoint === "" ? "Provide an endpoint name." : true,
        });
        if (config.rootConfig.endpoints[name] !== undefined) {
            const confirmed = await (0, prompts_1.confirm)({
                message: `The endpoint ${name} already exists. Overwrite?`,
                default: false,
            });
            if (!confirmed) {
                return this.askName(config);
            }
        }
        return name;
    }
    async askDefault(config, base, regions) {
        return (0, prompts_1.select)({
            message: "Endpoints created. Which endpoint would you like to set as default?",
            choices: [
                ...(config.rootConfig.defaultEndpoint !== undefined
                    ? [
                        {
                            name: `Keep '${config.rootConfig.defaultEndpoint}' endpoint as default`,
                            value: config.rootConfig.defaultEndpoint,
                        },
                    ]
                    : []),
                ...Object.values(regions).map((r) => ({ value: r.endpointName(base) })),
            ],
        });
    }
    async passwordStrategy() {
        return this.loginByPassword({
            email: await (0, prompts_1.input)({
                message: `Email address (from ${(0, chalk_1.underline)((0, chalk_1.blue)("https://dashboard.fauna.com/"))})`,
                validate: (email) => {
                    return !email || !/\S+@\S+\.\S+/.test(email)
                        ? "Provide a valid email address."
                        : true;
                },
            }),
            password: await (0, prompts_1.password)({
                message: "Password",
                mask: true,
            }),
        });
    }
    async otp({ email, password, }) {
        const otp = await (0, prompts_1.input)({
            message: "Enter your multi-factor authentication code",
        });
        return this.loginByPassword({
            email,
            password,
            otp,
        });
    }
    async handlePasswordStrategyError({ email, password, error, }) {
        if (["otp_required", "otp_invalid"].includes(error.code)) {
            if (error.code === "otp_invalid") {
                this.log(error.message);
            }
            return this.otp({ email, password });
        }
        if (error.code === "invalid_credentials") {
            this.log(error.message);
            return this.passwordStrategy();
        }
        throw error;
    }
    async loginByPassword({ email, password, otp, }) {
        const resp = await (0, node_fetch_1.default)(new URL("login", AUTH), {
            method: "POST",
            headers: { "Content-Type": "application/json" },
            body: JSON.stringify({
                email,
                password,
                type: "shell",
                session: "Fauna Shell - " + (0, os_1.hostname)(),
                ...(otp && { otp }),
            }),
        });
        const json = await resp.json();
        if (resp.ok) {
            return Object.fromEntries(Object.entries(json.regionGroups).map(([key, v]) => [
                key,
                new Region(key, v.secret),
            ]));
        }
        else {
            return this.handlePasswordStrategyError({ email, password, error: json });
        }
    }
}
CloudLoginCommand.description = "Log in to a Fauna account.";
CloudLoginCommand.examples = ["$ fauna cloud-login"];
CloudLoginCommand.flags = {};
exports.default = CloudLoginCommand;
