"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const prompts_1 = require("@inquirer/prompts");
const core_1 = require("@oclif/core");
const config_1 = require("../../lib/config");
const fauna_client_1 = tslib_1.__importDefault(require("../../lib/fauna-client"));
const secret_1 = require("../../lib/secret");
class AddEndpointCommand extends core_1.Command {
    async run() {
        const config = config_1.ShellConfig.read({});
        await this.execute(config);
    }
    async execute(config) {
        var _a, _b, _c;
        const { args, flags } = await this.parse();
        if (args.name === undefined && flags["non-interactive"]) {
            this.error("A name must be given if --non-interactive is set");
        }
        const endpointName = (_a = args === null || args === void 0 ? void 0 : args.name) !== null && _a !== void 0 ? _a : (await (0, prompts_1.input)({
            message: "Endpoint name",
            validate: (name) => {
                if (name === "default") {
                    return "Endpoint cannot be named 'default'";
                }
                else if (Object.keys(config.rootConfig.endpoints).includes(name)) {
                    return "Endpoint already exists";
                }
                else {
                    return true;
                }
            },
        }));
        const url = (_b = flags === null || flags === void 0 ? void 0 : flags.url) !== null && _b !== void 0 ? _b : (await (0, prompts_1.input)({
            message: "Database URL",
            default: "http://localhost:8443",
            validate: async (url) => {
                try {
                    // eslint-disable-next-line no-new
                    new URL(url);
                    return true;
                }
                catch (e) {
                    return "Error: invalid URL";
                }
            },
        }));
        let secret;
        if ((flags === null || flags === void 0 ? void 0 : flags.secret) === undefined) {
            const v = await (0, prompts_1.input)({
                message: "Database Secret",
                validate: (secret) => {
                    try {
                        secret_1.Secret.parse(secret);
                        return true;
                    }
                    catch (e) {
                        return e.message;
                    }
                },
            });
            secret = secret_1.Secret.parse(v);
        }
        else {
            secret = secret_1.Secret.parse(flags.secret);
        }
        core_1.ux.action.start("Checking secret");
        const client = new fauna_client_1.default({
            secret: secret.buildSecret(),
            endpoint: url,
        });
        try {
            const res = await client.query(`0`);
            if (res.status !== 200) {
                core_1.ux.action.stop();
                console.log("Warning: invalid secret");
            }
            else {
                core_1.ux.action.stop();
            }
        }
        catch (e) {
            core_1.ux.action.stop();
            console.log("Warning: could not connect to Fauna");
        }
        finally {
            await client.close();
        }
        const setDefault = (_c = flags === null || flags === void 0 ? void 0 : flags["set-default"]) !== null && _c !== void 0 ? _c : ((flags === null || flags === void 0 ? void 0 : flags["non-interactive"])
            ? false
            : await (0, prompts_1.confirm)({
                message: "Make this endpoint default",
            }));
        if (setDefault) {
            config.rootConfig.defaultEndpoint = endpointName;
        }
        config.rootConfig.endpoints[endpointName] = new config_1.Endpoint({
            secret,
            url,
        });
        config.saveRootConfig();
        this.log(`Saved endpoint ${endpointName} to ${(0, config_1.getRootConfigPath)()}`);
    }
}
AddEndpointCommand.args = {
    name: core_1.Args.string({
        description: "Endpoint name",
    }),
};
AddEndpointCommand.description = "Add an endpoint to ~/.fauna-shell.";
AddEndpointCommand.examples = [
    "$ fauna endpoint add",
    "$ fauna endpoint add localhost --url http://localhost:8443/ --key secret",
    "$ fauna endpoint add localhost --set-default",
];
AddEndpointCommand.flags = {
    url: core_1.Flags.string({
        description: "Database URL",
        required: false,
    }),
    secret: core_1.Flags.string({
        description: "Database secret",
        required: false,
    }),
    "non-interactive": core_1.Flags.boolean({
        description: "Disables interaction",
        dependsOn: ["url", "secret"],
    }),
    "set-default": core_1.Flags.boolean({
        description: "Sets this environment as the default",
    }),
};
AddEndpointCommand.aliases = ["add-endpoint"];
AddEndpointCommand.deprecateAliases = true;
exports.default = AddEndpointCommand;
