"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ProjectInitCommand = void 0;
const tslib_1 = require("tslib");
const prompts_1 = require("@inquirer/prompts");
const core_1 = require("@oclif/core");
const fs_1 = tslib_1.__importDefault(require("fs"));
const path = tslib_1.__importStar(require("path"));
const config_1 = require("../../lib/config");
const environment_factory_1 = require("../../lib/environment-factory");
const file_util_1 = require("../../lib/file-util");
class ProjectInitCommand extends core_1.Command {
    async run() {
        const { args } = await this.parse();
        const projectDir = this.getProjectPath(args.projectDir);
        await this.execute(projectDir);
    }
    async execute(projectDir) {
        if (!(0, file_util_1.dirExists)(projectDir)) {
            this.log(`${projectDir} does not exist, creating it.`);
            fs_1.default.mkdirSync(projectDir, { recursive: true });
        }
        else if (!(0, file_util_1.dirIsWriteable)(projectDir)) {
            this.error(`${projectDir} is not writeable.`);
        }
        const projectPath = path.join(projectDir, config_1.PROJECT_FILE_NAME);
        if ((0, config_1.fileExists)(projectPath)) {
            this.error(`Project already exists at ${projectPath}`);
        }
        this.log(`Creating project at ${projectPath}`);
        const existingProject = (0, config_1.getProjectConfigPath)(projectPath);
        if (existingProject !== undefined) {
            this.log(`Warning: the new project will override the existing project at ${existingProject}`);
        }
        let schemaDir = await (0, prompts_1.input)({
            message: "What directory would you like to store your schema files in? (defaults to current directory)",
        });
        const fullSchemaPath = path.join(projectDir, schemaDir);
        if (schemaDir === "") {
            schemaDir = undefined;
        }
        else if (!(0, file_util_1.dirExists)(fullSchemaPath)) {
            this.log(`The project's schema directory: ${fullSchemaPath} does not exist, creating it.`);
            fs_1.default.mkdirSync(fullSchemaPath, { recursive: true });
        }
        else if (!(0, file_util_1.dirIsWriteable)(fullSchemaPath)) {
            this.error(`The project's schema directory: ${fullSchemaPath} is not writeable.`);
        }
        config_1.ProjectConfig.initialConfig(schemaDir).save(projectPath);
        const shellConfig = config_1.ShellConfig.readWithOverrides({
            projectPath: projectPath,
        });
        const environmentFactory = new environment_factory_1.EnvironmentFactory(this, shellConfig);
        await environmentFactory.addEnvironment({
            default: true,
        });
    }
    getProjectPath(projectDir) {
        if (projectDir) {
            if (path.isAbsolute(projectDir)) {
                return projectDir;
            }
            else {
                return path.join(process.cwd(), projectDir);
            }
        }
        else {
            return process.cwd();
        }
    }
}
exports.ProjectInitCommand = ProjectInitCommand;
ProjectInitCommand.args = {
    projectDir: core_1.Args.string({
        name: "projectDir",
        required: false,
        description: "The directory to initialize as a fauna project.  If not provided will default to the current directory.",
    }),
};
ProjectInitCommand.description = `Initialize a project directory by generating a .fauna-project file.

NOTE: \`fauna project\` and \`fauna environment\` are still in beta. Behavior is subject to change.`;
ProjectInitCommand.examples = [
    "$ fauna project init",
    "$ fauna project init path/to/some/other/dir",
];
