"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const prompts_1 = require("@inquirer/prompts");
const schema_command_1 = tslib_1.__importDefault(require("../../lib/schema-command"));
const node_fetch_1 = tslib_1.__importDefault(require("node-fetch"));
const fs_1 = tslib_1.__importDefault(require("fs"));
const path_1 = tslib_1.__importDefault(require("path"));
const core_1 = require("@oclif/core");
class PullSchemaCommand extends schema_command_1.default {
    async run() {
        var _a, _b;
        const { url, secret } = await this.fetchsetup();
        try {
            // Gather remote schema files to download.
            const filesres = await (0, node_fetch_1.default)(new URL("/schema/1/files", url), {
                method: "GET",
                headers: { AUTHORIZATION: `Bearer ${secret}` },
            });
            const filesjson = await filesres.json();
            if (filesjson.error) {
                this.error(filesjson.error.message);
            }
            // Sort for consistent order. It's nice for tests.
            const filenames = filesjson.files
                .map((file) => file.filename)
                .filter((name) => name.endsWith(".fsl"))
                .sort();
            // Gather local .fsl files to overwrite or delete.
            const existing = this.gatherRelativeFSLFilePaths();
            // Summarize file changes.
            const adds = [];
            const overwrites = [];
            for (const fn of filenames) {
                if (existing.includes(fn)) {
                    overwrites.push(fn);
                }
                else {
                    adds.push(fn);
                }
            }
            const deletes = [];
            for (const fn of existing) {
                if (!filenames.includes(fn)) {
                    deletes.push(fn);
                }
            }
            deletes.sort();
            console.log("Pull makes the following changes:");
            if ((_a = this.flags) === null || _a === void 0 ? void 0 : _a.delete) {
                for (const deleteme of deletes) {
                    console.log(`delete:    ${deleteme}`);
                }
            }
            for (const add of adds) {
                console.log(`add:       ${add}`);
            }
            for (const overwrite of overwrites) {
                console.log(`overwrite: ${overwrite}`);
            }
            if ((_b = this.flags) === null || _b === void 0 ? void 0 : _b.delete) {
                // Delete extra .fsl files.
                for (const deleteme of deletes) {
                    fs_1.default.unlinkSync(path_1.default.join(this.dir, deleteme));
                }
            }
            const confirmed = await (0, prompts_1.confirm)({
                message: "Accept the changes?",
                default: false,
            });
            if (confirmed) {
                for (const filename of filenames) {
                    const fileres = await (0, node_fetch_1.default)(new URL(`/schema/1/files/${encodeURIComponent(filename)}`, url), {
                        method: "GET",
                        headers: { AUTHORIZATION: `Bearer ${secret}` },
                    });
                    const filejson = await fileres.json();
                    if (filejson.error) {
                        this.error(filejson.error.message);
                    }
                    const fp = path_1.default.join(this.dir, filename);
                    fs_1.default.mkdirSync(path_1.default.dirname(fp), { recursive: true });
                    fs_1.default.writeFileSync(fp, filejson.content);
                }
            }
            else {
                this.log("Change cancelled");
            }
        }
        catch (err) {
            this.error(err);
        }
    }
}
PullSchemaCommand.flags = {
    ...schema_command_1.default.flags,
    delete: core_1.Flags.boolean({
        description: "Delete .fsl files in the target directory that are not part of the database schema",
        default: false,
    }),
};
PullSchemaCommand.description = "Pull a database schema's .fsl files into the current project.";
PullSchemaCommand.examples = ["$ fauna schema pull"];
exports.default = PullSchemaCommand;
