"use strict";
// Manages ~/.fauna-shell
Object.defineProperty(exports, "__esModule", { value: true });
exports.fileExists = exports.getProjectConfigPath = exports.getRootConfigPath = exports.ShellConfig = exports.Config = exports.InvalidConfigError = exports.PROJECT_FILE_NAME = exports.RootConfig = exports.ProjectConfig = exports.Environment = exports.Endpoint = void 0;
const tslib_1 = require("tslib");
const fs_1 = tslib_1.__importDefault(require("fs"));
const os_1 = tslib_1.__importDefault(require("os"));
const path_1 = tslib_1.__importDefault(require("path"));
const ini = require("ini");
const secret_1 = require("../secret");
const project_config_1 = require("./project-config");
Object.defineProperty(exports, "Environment", { enumerable: true, get: function () { return project_config_1.Environment; } });
Object.defineProperty(exports, "ProjectConfig", { enumerable: true, get: function () { return project_config_1.ProjectConfig; } });
const root_config_1 = require("./root-config");
Object.defineProperty(exports, "Endpoint", { enumerable: true, get: function () { return root_config_1.Endpoint; } });
Object.defineProperty(exports, "RootConfig", { enumerable: true, get: function () { return root_config_1.RootConfig; } });
exports.PROJECT_FILE_NAME = ".fauna-project";
class InvalidConfigError extends Error {
}
exports.InvalidConfigError = InvalidConfigError;
// Wraps an `ini` file with helpers to get typed values out
class Config {
    constructor(keyName, config) {
        this.keyName = keyName;
        this.config = config;
    }
    strOpt(key) {
        const v = this.config[key];
        if (v === undefined || typeof v === "string") {
            return v;
        }
        else {
            throw new InvalidConfigError(`Expected string for ${this.keyName} ${key}, got ${typeof v}`);
        }
    }
    keys() {
        return Object.keys(this.config);
    }
    numberOpt(key) {
        const v = this.config[key];
        if (v === undefined || typeof v === "number") {
            return v;
        }
        else if (typeof v === "string") {
            try {
                return parseInt(v, 10);
            }
            catch (_) {
                throw new InvalidConfigError(`Invalid number for ${this.keyName} ${key}`);
            }
        }
        else {
            throw new InvalidConfigError(`Expected number for ${this.keyName} ${key}, got ${typeof v}`);
        }
    }
    object(key) {
        const v = this.config[key];
        if (v === undefined) {
            throw new InvalidConfigError(`Missing value for required ${this.keyName} ${key}`);
        }
        else if (v !== null && typeof v === "object") {
            return new Config(this.keyName, v);
        }
        else {
            throw new InvalidConfigError(`Expected object for ${this.keyName} ${key}, got ${typeof v}`);
        }
    }
    objectExists(key) {
        const v = this.config[key];
        return v !== undefined && typeof v === "object";
    }
    str(key) {
        return this.require(key, this.strOpt(key));
    }
    number(key) {
        return this.require(key, this.numberOpt(key));
    }
    require(key, value) {
        if (value === undefined) {
            throw new InvalidConfigError(`Missing value for required ${this.keyName} ${key}`);
        }
        else {
            return value;
        }
    }
    // Returns a list of all keys that match `pred`.
    allObjectsWhere(pred) {
        return Object.keys(this.config).flatMap((k) => pred(k) ? [[k, this.object(k)]] : []);
    }
    // Returns a list of all child objects of the given `key`.
    objectsIn(key) {
        const obj = this.object(key);
        return obj.allObjectsWhere((_) => true);
    }
}
exports.Config = Config;
class ShellConfig {
    static read(flags, log) {
        const rootConfig = ini.parse(readFileOpt((0, exports.getRootConfigPath)()));
        const projectConfigPath = (0, exports.getProjectConfigPath)();
        const projectConfig = projectConfigPath
            ? ini.parse(readFile(projectConfigPath))
            : undefined;
        const shellConfig = new ShellConfig({
            flags,
            rootConfig,
            projectPath: projectConfigPath !== undefined
                ? path_1.default.dirname(projectConfigPath)
                : undefined,
            projectConfig,
        });
        if (log !== undefined) {
            shellConfig.configErrors().forEach((err) => log.warn(err));
        }
        return shellConfig;
    }
    static readWithOverrides(opts) {
        var _a, _b, _c;
        const rootConfig = (_a = opts === null || opts === void 0 ? void 0 : opts.rootConfig) !== null && _a !== void 0 ? _a : ini.parse(readFileOpt((0, exports.getRootConfigPath)()));
        const projectConfigPath = (_b = opts === null || opts === void 0 ? void 0 : opts.projectPath) !== null && _b !== void 0 ? _b : (0, exports.getProjectConfigPath)();
        const projectConfig = (_c = opts === null || opts === void 0 ? void 0 : opts.projectConfig) !== null && _c !== void 0 ? _c : (projectConfigPath ? ini.parse(readFile(projectConfigPath)) : undefined);
        return new ShellConfig({
            rootConfig,
            projectPath: projectConfigPath !== undefined
                ? path_1.default.dirname(projectConfigPath)
                : undefined,
            projectConfig,
        });
    }
    constructor(opts) {
        var _a, _b, _c, _d, _e, _f, _g;
        // Errors that came up in the config to be displayed.
        this.errors = [];
        this.validate = () => {
            var _a;
            if (this.endpoint === undefined && this.secretFlag === undefined) {
                // No `~/.fauna-shell` was found, and no `--secret` was passed.
                throw new Error(`No endpoint or secret set. Set an endpoint in ~/.fauna-shell, ${exports.PROJECT_FILE_NAME}, or pass --endpoint`);
                /**
                 * If there is no secret flag set we need to ensure we validate we can find a secret
                 * from the endpoint.  Additionally if there is a root config present, we
                 * want to validate that things line up with the project.  Even if a secret
                 * flag is set, there could be other properties of the endpoint that we need to
                 * pull in, url being the current one.
                 * The inverse of this is the running from a pipeline scenario where there is a secret
                 * set and no root config.  In that case we don't want to validate the project configuration.
                 */
            }
            else if (this.secretFlag === undefined || !this.rootConfig.isEmpty()) {
                (_a = this.projectConfig) === null || _a === void 0 ? void 0 : _a.validate(this.rootConfig);
            }
        };
        this.lookupEndpoint = (opts) => {
            var _a, _b;
            this.validate();
            if (this.secretFlag !== undefined) {
                const endpoint = this.endpoint.makeScopedEndpoint();
                endpoint.secret = this.secretFlag;
                return endpoint;
            }
            else {
                let database = (_b = (_a = this.environment) === null || _a === void 0 ? void 0 : _a.database.split("/")) !== null && _b !== void 0 ? _b : [];
                if (opts.scope !== undefined) {
                    database.push(...opts.scope.split("/"));
                }
                return this.endpoint.makeScopedEndpoint(database);
            }
        };
        this.flags = new Config("flag", (_a = opts.flags) !== null && _a !== void 0 ? _a : {});
        this.rootConfig = new root_config_1.RootConfig(new Config("config key", (_b = opts.rootConfig) !== null && _b !== void 0 ? _b : {}));
        this.projectPath = opts.projectPath;
        this.projectConfig = opts.projectConfig
            ? project_config_1.ProjectConfig.fromConfig(new Config("config key", opts.projectConfig))
            : undefined;
        const urlFlag = root_config_1.Endpoint.getURLFromFlags(this.flags);
        if (urlFlag !== undefined) {
            try {
                // eslint-disable-next-line no-new
                new URL(urlFlag);
            }
            catch (e) {
                throw new Error(`Invalid database URL: ${urlFlag}`);
            }
        }
        const environmentFlag = this.flags.strOpt("environment");
        if (this.projectConfig === undefined) {
            if (environmentFlag !== undefined) {
                throw new Error(`No ${exports.PROJECT_FILE_NAME} was found, so environment '${environmentFlag}' cannot be used`);
            }
        }
        else {
            const environmentName = environmentFlag !== null && environmentFlag !== void 0 ? environmentFlag : this.projectConfig.defaultEnvironment;
            if (environmentName !== undefined) {
                this.environment = this.projectConfig.environments[environmentName];
                // we don't need to blow up here because commands that aren't using project config
                // can hit this.  cloud-login could be trying to add the needed endpoint.
                // the project config will also be validated for any commands attempting to use it.
                if (this.environment === undefined) {
                    this.errors.push(`No such environment '${environmentName}'`);
                }
            }
        }
        // An endpoint must be chosen as well. The endpoint may come from (in order)
        // the `--endpoint` flag, the `endpoint` key in the project config, or the
        // `default` key in the root config.
        const endpointName = (_e = (_c = this.flags.strOpt("endpoint")) !== null && _c !== void 0 ? _c : (_d = this.environment) === null || _d === void 0 ? void 0 : _d.endpoint) !== null && _e !== void 0 ? _e : this.rootConfig.defaultEndpoint;
        const secretFlag = this.flags.strOpt("secret");
        this.secretFlag =
            secretFlag !== undefined ? secret_1.Secret.parseFlag(secretFlag) : undefined;
        if (secretFlag !== undefined && environmentFlag !== undefined) {
            throw new Error("Cannot specify both --secret and --environment, as --secret will override the settings from a environment.");
        }
        // There are 2 scenarios where we want to execute the first if block:
        // 1. no endpoint name is set
        // 2. In a CI/CD environment where a secret is set but there is no
        // root config.  In this scenario endpoints may be set by the project
        // configuration but won't exist in their pipeline workspace.
        if (endpointName === undefined ||
            (secretFlag !== undefined && this.rootConfig.isEmpty())) {
            // This is a dummy secret. `--secret` must be set in this case, which
            // `validate` enforces.
            this.endpoint = new root_config_1.Endpoint({
                secret: new secret_1.Secret({ key: "", allowDatabase: true }),
                url: urlFlag,
                graphqlHost: this.flags.strOpt("graphqlHost"),
                graphqlPort: this.flags.numberOpt("graphqlPort"),
            });
        }
        else {
            this.endpoint = this.rootConfig.endpoints[endpointName];
            if (this.endpoint !== undefined) {
                this.endpoint.url = urlFlag !== null && urlFlag !== void 0 ? urlFlag : this.endpoint.url;
                this.endpoint.graphqlHost =
                    (_f = this.flags.strOpt("graphqlHost")) !== null && _f !== void 0 ? _f : this.endpoint.graphqlHost;
                this.endpoint.graphqlPort =
                    (_g = this.flags.numberOpt("graphqlHost")) !== null && _g !== void 0 ? _g : this.endpoint.graphqlPort;
            }
        }
    }
    configErrors() {
        if (this.rootConfig.invalidEndpoints.length > 0) {
            return [
                `The following endpoint definitions in ${(0, exports.getRootConfigPath)()} are invalid:\n ${this.rootConfig.invalidEndpoints.join("\n")}\n Resolve them by ensuring they have a secret defined or remove them if they are not needed.`,
                ...this.errors,
            ];
        }
        else {
            return [];
        }
    }
    /**
     * Saves the project config, if present.
     */
    saveProjectConfig() {
        var _a;
        (_a = this.projectConfig) === null || _a === void 0 ? void 0 : _a.save(this.projectConfigFile());
    }
    /**
     * Saves the root config.
     */
    saveRootConfig() {
        if (this.rootConfig.invalidEndpoints.length > 0) {
            throw new Error(`The following endpoint definitions in ${(0, exports.getRootConfigPath)()} are invalid:\n ${this.rootConfig.invalidEndpoints.join("\n")}\n Resolve them by ensuring they have a secret defined or remove them if they are not needed.`);
        }
        this.rootConfig.save((0, exports.getRootConfigPath)());
    }
    projectConfigFile() {
        return this.projectPath === undefined
            ? undefined
            : path_1.default.join(this.projectPath, exports.PROJECT_FILE_NAME);
    }
}
exports.ShellConfig = ShellConfig;
const readFileOpt = (fileName) => {
    if (fs_1.default.existsSync(fileName)) {
        return fs_1.default.readFileSync(fileName, "utf8");
    }
    else {
        return "";
    }
};
const readFile = (fileName) => {
    return fs_1.default.readFileSync(fileName, "utf8");
};
const getRootConfigPath = () => {
    return path_1.default.join(os_1.default.homedir(), ".fauna-shell");
};
exports.getRootConfigPath = getRootConfigPath;
const getProjectConfigPath = (start) => {
    let current = start;
    if (current === undefined) {
        try {
            current = process.cwd();
        }
        catch (err) {
            // If the cwd is not accessible, just give up. If this happens, one of our
            // dependencies actually explodes elsewhere, but we might as well handle
            // errors where possible.
            return undefined;
        }
    }
    // eslint-disable-next-line no-constant-condition
    while (true) {
        const projPath = path_1.default.join(current, exports.PROJECT_FILE_NAME);
        if ((0, exports.fileExists)(projPath)) {
            return projPath;
        }
        const currPath = path_1.default.parse(current);
        // break if at root
        if (currPath.base === "") {
            break;
        }
        current = currPath.dir;
    }
    // if we got here, it means that there was no `.fauna-project` file, so we
    // give up.
    return undefined;
};
exports.getProjectConfigPath = getProjectConfigPath;
const fileExists = (filePath) => {
    const stat = fs_1.default.statSync(filePath, {
        // returns undefined instead of throwing if the file doesn't exist
        throwIfNoEntry: false,
    });
    return stat !== undefined && stat.isFile();
};
exports.fileExists = fileExists;
