"use strict";
var _FaunaObjectTranslator_instances, _a, _FaunaObjectTranslator_NUMBER_REGEX, _FaunaObjectTranslator_TRULY, _FaunaObjectTranslator_getErrorValue, _FaunaObjectTranslator_typeCasting, _FaunaObjectTranslator_logger, _FaunaObjectTranslator_getNumber, _FaunaObjectTranslator_stringBool, _FaunaObjectTranslator_stringDate, _FaunaObjectTranslator_epochMillisDate, _FaunaObjectTranslator_epochSecondsDate, _FaunaObjectTranslator_prepareRecord, _FaunaObjectTranslator_castType;
Object.defineProperty(exports, "__esModule", { value: true });
exports.FaunaObjectTranslator = exports.TranslationError = void 0;
const tslib_1 = require("tslib");
const faunadb_1 = require("faunadb");
const moment_1 = require("moment");
/**
 * An error translating an object with a {FaunaObjectTranslater}.
 */
class TranslationError extends Error {
}
exports.TranslationError = TranslationError;
/**
 * Helper class for cleaning objects prior to persistence in Fauna.
 * It has two main objectives:
 *   - trim input strings to remove unneeded whitespace
 *   - cast types as specified by input
 **/
class FaunaObjectTranslator {
    /**
     * Constructs a new FaunaObjectTranslator
     * @param {Array<string>} typeTranslations - custom typeTranslations to perform on object
     *   keys. Must be provided as an array of strings in the form:
     *      [keyName::typeTranslationFunction, ...]
     *   Supported typeTranslationFunctions are: number, dateString, dateEpochMills, dateEpochSeconds, bool
     *   e.g. [myKey::bool, myOtherKey::number, myFinalKey::date_string]
     */
    constructor(typeTranslations, logger = console.log) {
        _FaunaObjectTranslator_instances.add(this);
        _FaunaObjectTranslator_typeCasting.set(this, void 0);
        _FaunaObjectTranslator_logger.set(this, void 0);
        tslib_1.__classPrivateFieldSet(this, _FaunaObjectTranslator_logger, logger, "f");
        tslib_1.__classPrivateFieldSet(this, _FaunaObjectTranslator_typeCasting, (() => {
            if (!typeTranslations)
                return {};
            const colTypeCast = {
                number: tslib_1.__classPrivateFieldGet(this, _FaunaObjectTranslator_instances, "m", _FaunaObjectTranslator_getNumber),
                dateString: tslib_1.__classPrivateFieldGet(this, _FaunaObjectTranslator_instances, "m", _FaunaObjectTranslator_stringDate),
                dateEpochMillis: tslib_1.__classPrivateFieldGet(this, _FaunaObjectTranslator_instances, "m", _FaunaObjectTranslator_epochMillisDate),
                dateEpochSeconds: tslib_1.__classPrivateFieldGet(this, _FaunaObjectTranslator_instances, "m", _FaunaObjectTranslator_epochSecondsDate),
                bool: tslib_1.__classPrivateFieldGet(this, _FaunaObjectTranslator_instances, "m", _FaunaObjectTranslator_stringBool),
            };
            const types = typeTranslations.reduce((memo, next) => {
                const indexOfTranslationTag = next.lastIndexOf("::");
                const [name, typeTranslator] = [
                    next.substring(0, indexOfTranslationTag),
                    next.substring(indexOfTranslationTag + 2, next.length),
                ];
                return {
                    casting: {
                        ...memo.casting,
                        [name]: { typeTranslator, castFn: colTypeCast[typeTranslator] },
                    },
                    invalidType: colTypeCast[typeTranslator]
                        ? memo.invalidType
                        : [...memo.invalidType, name],
                };
            }, { casting: {}, invalidType: [] });
            if (types.invalidType.length !== 0) {
                throw new Error(`The following columns have an invalid type translation specified: ${types.invalidType}`);
            }
            return types.casting;
        })(), "f");
    }
    /**
     * Translates the rawData to a cleaned Fauna object - applying
     * any type transformations and trimming field names.
     * @param {object} rawData - the uncleaned record
     * @return {object} An object holding the cleaned data.
     * @throws {Error} if translation fails
     */
    getRecord(rawData) {
        return tslib_1.__classPrivateFieldGet(this, _FaunaObjectTranslator_instances, "m", _FaunaObjectTranslator_castType).call(this, tslib_1.__classPrivateFieldGet(this, _FaunaObjectTranslator_instances, "m", _FaunaObjectTranslator_prepareRecord).call(this, rawData));
    }
}
exports.FaunaObjectTranslator = FaunaObjectTranslator;
_a = FaunaObjectTranslator, _FaunaObjectTranslator_typeCasting = new WeakMap(), _FaunaObjectTranslator_logger = new WeakMap(), _FaunaObjectTranslator_instances = new WeakSet(), _FaunaObjectTranslator_getErrorValue = function _FaunaObjectTranslator_getErrorValue(val) {
    return typeof val === "string" ? val : JSON.stringify(val);
}, _FaunaObjectTranslator_getNumber = function _FaunaObjectTranslator_getNumber(val) {
    if (typeof val === "number") {
        return val;
    }
    if (typeof val === "string" &&
        tslib_1.__classPrivateFieldGet(_a, _a, "f", _FaunaObjectTranslator_NUMBER_REGEX).test(val)) {
        return Number(val);
    }
    throw new TranslationError(`Invalid number '${tslib_1.__classPrivateFieldGet(_a, _a, "m", _FaunaObjectTranslator_getErrorValue).call(_a, val)}' cannot be translated to a number.`);
}, _FaunaObjectTranslator_stringBool = function _FaunaObjectTranslator_stringBool(val) {
    return (tslib_1.__classPrivateFieldGet(_a, _a, "f", _FaunaObjectTranslator_TRULY).includes(val) ||
        (typeof val === "string" &&
            tslib_1.__classPrivateFieldGet(_a, _a, "f", _FaunaObjectTranslator_TRULY).includes(val.toLowerCase())));
}, _FaunaObjectTranslator_stringDate = function _FaunaObjectTranslator_stringDate(val) {
    if (typeof val !== "string") {
        throw new TranslationError(`Cannot convert '${tslib_1.__classPrivateFieldGet(_a, _a, "m", _FaunaObjectTranslator_getErrorValue).call(_a, val)}' to a date.`);
    }
    let theDate = (0, moment_1.utc)(val, moment_1.ISO_8601);
    if (!theDate.isValid()) {
        // fallback to other date formats
        theDate = (0, moment_1.utc)(val, moment_1.RFC_2822);
        if (!theDate.isValid()) {
            theDate = new Date(val);
            if (Number.isNaN(theDate.getTime())) {
                throw new TranslationError(`The string '${val}' cannot be translated to a date.`);
            }
            tslib_1.__classPrivateFieldGet(this, _FaunaObjectTranslator_logger, "f").call(this, `Warning: the string '${val}' is not valid ISO-8601 nor RFC_2822 date. \
Making a best-effort translation to '${theDate}'`);
        }
    }
    return faunadb_1.query.Time(theDate.toISOString());
}, _FaunaObjectTranslator_epochMillisDate = function _FaunaObjectTranslator_epochMillisDate(val) {
    try {
        return faunadb_1.query.Time((0, moment_1.unix)(tslib_1.__classPrivateFieldGet(this, _FaunaObjectTranslator_instances, "m", _FaunaObjectTranslator_getNumber).call(this, val) / 1000)
            .utc()
            .toISOString());
    }
    catch (e) {
        throw new TranslationError(`Cannot convert '${tslib_1.__classPrivateFieldGet(_a, _a, "m", _FaunaObjectTranslator_getErrorValue).call(_a, val)}' to a date.`);
    }
}, _FaunaObjectTranslator_epochSecondsDate = function _FaunaObjectTranslator_epochSecondsDate(val) {
    try {
        return faunadb_1.query.Time((0, moment_1.unix)(tslib_1.__classPrivateFieldGet(this, _FaunaObjectTranslator_instances, "m", _FaunaObjectTranslator_getNumber).call(this, val)).utc().toISOString());
    }
    catch (e) {
        throw new TranslationError(`Cannot convert '${tslib_1.__classPrivateFieldGet(_a, _a, "m", _FaunaObjectTranslator_getErrorValue).call(_a, val)}' to a date.`);
    }
}, _FaunaObjectTranslator_prepareRecord = function _FaunaObjectTranslator_prepareRecord(obj) {
    return Object.keys(obj).reduce((memo, next) => {
        memo[next.trim()] = obj[next];
        return memo;
    }, {});
}, _FaunaObjectTranslator_castType = function _FaunaObjectTranslator_castType(obj) {
    return Object.keys(tslib_1.__classPrivateFieldGet(this, _FaunaObjectTranslator_typeCasting, "f")).reduce((memo, col) => {
        if (memo[col] === undefined || memo[col] === null)
            return memo;
        const castedValue = tslib_1.__classPrivateFieldGet(this, _FaunaObjectTranslator_typeCasting, "f")[col].castFn.call(this, memo[col]);
        if (castedValue !== undefined) {
            memo[col] = castedValue;
        }
        else {
            tslib_1.__classPrivateFieldGet(this, _FaunaObjectTranslator_logger, "f").call(this, `Value '${memo[col]}' at column '${col}' can not be casted to type '${tslib_1.__classPrivateFieldGet(this, _FaunaObjectTranslator_typeCasting, "f")[col].type}'`);
        }
        return memo;
    }, obj);
};
_FaunaObjectTranslator_NUMBER_REGEX = { value: /(^\s*[+|-]?\d+\s*$)|(^\s*[+|-]?\d*\.\d+\s*$)/ };
_FaunaObjectTranslator_TRULY = { value: ["true", "t", "yes", "1", 1, true] };
