"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.searchSelect = void 0;
const tslib_1 = require("tslib");
const core_1 = require("@inquirer/core");
const chalk_1 = tslib_1.__importDefault(require("chalk"));
const figures_1 = tslib_1.__importDefault(require("figures"));
function renderItem({ item, isActive }) {
    const line = item.value;
    const prefix = isActive ? figures_1.default.pointer : ` `;
    const str = `${prefix} ${line}`;
    return isActive ? chalk_1.default.cyan(str) : str;
}
// @inquirer/core has these, but it includes the vim keybinds j and k, which
// conflict with typing those letters.
const isUpKey = (key) => key.name === "up" || // The up key
    (key.ctrl && key.name === "p"); // Emacs keybinding
const isDownKey = (key) => key.name === "down" || // The down key
    (key.ctrl && key.name === "n"); // Emacs keybinding
exports.searchSelect = (0, core_1.createPrompt)((config, done) => {
    var _a;
    const { choices } = config;
    const pageSize = (_a = config.pageSize) !== null && _a !== void 0 ? _a : 7;
    const firstRender = (0, core_1.useRef)(true);
    const prefix = (0, core_1.usePrefix)();
    const [status, setStatus] = (0, core_1.useState)("pending");
    const [typed, setTyped] = (0, core_1.useState)("");
    // TODO: Fuzzy filter?
    const items = choices.filter((item, _) => item.value.startsWith(typed));
    // This is the item with a cursor next to it. The arrow keys will
    // increment/decrement this value.
    const [activeState, setActive] = (0, core_1.useState)(0);
    // This is the distance the amount that has been scrolled down. It is the
    // index of the first visible item.
    const [scrollState, setScroll] = (0, core_1.useState)(0);
    // `typed` can get updated, changing the length of `items`, without `active`
    // or `scroll` being updated. Its simplest just to truncate these values
    // here.
    const active = Math.min(activeState, items.length - 1);
    const scroll = Math.min(scrollState, items.length - pageSize);
    // Now that `active` has been truncated, we can just index into this without
    // worrying about undefined.
    const selectedChoice = items[active];
    (0, core_1.useKeypress)((key, rl) => {
        if ((0, core_1.isEnterKey)(key)) {
            if (items.length > 0) {
                setStatus("done");
                done(selectedChoice.value);
            }
        }
        else if (isUpKey(key) || isDownKey(key)) {
            const offset = isUpKey(key) ? -1 : 1;
            let next = Math.min(Math.max(0, active + offset), items.length - 1);
            setActive(next);
            if (next < scroll) {
                setScroll(next);
            }
            else if (next >= scroll + pageSize) {
                setScroll(next - pageSize + 1);
            }
        }
        else if ((0, core_1.isBackspaceKey)(key)) {
            if (typed.length > 0) {
                setTyped(typed.slice(0, typed.length - 1));
            }
        }
        else if (!key.ctrl) {
            setTyped(rl.line);
        }
    });
    let message = chalk_1.default.bold(config.message);
    if (firstRender.current) {
        firstRender.current = false;
        message += chalk_1.default.dim(" (Type to filter, arrow keys to select)");
    }
    if (status === "done") {
        return `${prefix} ${message} ${chalk_1.default.cyan(selectedChoice.value)}`;
    }
    const lines = items.map((item, index) => renderItem({ item, isActive: index === active }));
    let page;
    if (lines.length === 0) {
        page = chalk_1.default.dim("  <No results>");
    }
    else if (lines.length > pageSize) {
        page = lines.slice(scroll, scroll + pageSize).join("\n");
    }
    else {
        page = lines.join("\n");
    }
    // Return value is an array of two elements, with the cursor rendered
    // between the two.
    return [`${prefix} ${message} ${typed}`, page];
});
